// Programmer and reader for DS1225Y and RAMTRON 18W08-SG*
// Howardlong, 9 Jan 2015
//
// Connections are:
// RD0:RD7 -> DQ0:DQ7
// RB0:RB5 -> A0:A5
// RA0:RA6 -> A6:A12
// RE0 -> \WE
// RE1 -> \OE
// RE2 -> \CE
// RC6 -> TXD to terminal emulator
// RC7 -> RXD from terminal emulator
//
// Device data is read into PIC's flash program memory, so it is kept during power off.
//
// Procedure:
//
// o Setup a terminal emulator for 115200 n 8 1 and connect up
// o Insert source NVRAM
// o Power on (+5V)
// o [R]ead NVRAM into PIC flash (verify is also performed)
// o get [C]hecksums and take a note of them, they should be the same
// o Power off
// o Replace NVRAM to new NVRAM
// o Power on (+5V)
// o get [C]hecksums and check that the Flash checksum is the same
// o [W]rite PIC flash to NVRAM (verify is also performed)
// o get [C]hecksums and take a note of them, they should all be the same
// o Power off
//
// Depending on the version of XC8 compiler used (V1.32A has been tested) you may see these advisories and warnings which may be ignored:
//
// o Advisory[1233] Employing 18F4550 errata work-arounds:
// o Advisory[1234]  * Corrupted fast interrupt shadow registers
// o Warning [1467] D:\Program Files (x86)\Microchip\xc8\v1.32\sources\pic18\plib\Flash\WriteBlockFlash.c; 75. read-only target "ROM" may be indirectly written via pointer
//
// Note that the serial send is throttled back by introducing a 200us delay per character, as Putty seems not to be able to cope with large amounts of
// continuously streamed information at 115kbps.
//
// * Note that the RAMTRON 18W08-SG is a 28 pin SOIC so needs to be placed on a DIL adapter to replace the DS1225Y. In addition, the RAMTRON device has
// two additional address lines, on pins 1 and 26, that need to be tied for ground: this programmer only deals with the first 8k bytes (i.e., same size as the
// DS1225Y).
//

#include <xc.h>
#include <stdint.h>
#include <stdlib.h>
#include <ctype.h>

#include "main.h"

#pragma config CONFIG1L=0x00 // Set to avoid warning 1311 about missing config setting for 0x300000
#pragma config CONFIG6H=0xE0 // Set to avoid warning 1311 about missing config setting for 0x30000B
#pragma config FOSC=INTOSCIO_EC
#pragma config WDT=OFF
#pragma config BOR=ON
#pragma config BORV=1 // brown out 4.11-4.55V
#pragma config PWRT=ON
#pragma config MCLRE=ON
#pragma config LVP=OFF
#ifdef __DEBUG
	#pragma config WRT0=OFF
	#pragma config WRT2=OFF
	#pragma config WRT3=OFF
#else
	#pragma config WRT0=ON
	#pragma config WRT2=ON
	#pragma config WRT3=ON
#endif
#pragma config WRT1=OFF // Block 1 is the flash write block for a copy of the NVRAM. Note block 3 is used by debug executive.

#define FLASH_WRITE_SIZE (32)
#define MAX_VERIFY_ERRORS (10)

#define TRIS_nCE TRISEbits.TRISE2
#define LAT_nCE LATEbits.LATE2
#define TRIS_nOE TRISEbits.TRISE1
#define LAT_nOE LATEbits.LATE1
#define TRIS_nWE TRISEbits.TRISE0
#define LAT_nWE LATEbits.LATE0

extern const U8 _au8Flash[NVRAM_SIZE];

static void UARTInit(U32 lBaudRate)
{
	U32 ul;
	U16 us;
	
	ul=4*lBaudRate;
	ul=_XTAL_FREQ/ul;
	us=(U16)(ul-1);
	
	SPBRG=(U8)us;                                   //Writing SPBRG Register
	SPBRGH=(U8)(us>>8);
	
	TRISCbits.TRISC6=1;
	TRISCbits.TRISC7=1;
	
	TXSTAbits.TX9=1;
	TXSTAbits.TXEN=1;
	TXSTAbits.SYNC=0;
	TXSTAbits.BRGH=1;
	TXSTAbits.TX9D=1;
	RCSTAbits.CREN=1;
	BAUDCONbits.BRG16=1;
	RCSTAbits.SPEN=1;
}

static void UARTWrite(char c)
{
	__delay_us(200); // Putty seems to miss a bit if we don't put a delay in.					
	while(!TXSTAbits.TRMT)
	{
		NOP();
	}
  	TXREG=c;
}

static void UARTWriteText(const char *psz)
{
	char c;
	
	while ((c=*psz++)!='\0')
	{
		UARTWrite(c);
	}
}

static void UARTWriteHexNibble(U8 u8)
{
	static const char acHex[]="0123456789ABCDEF";
	
	UARTWrite(acHex[u8 & 0xF]);
}
	
static void UARTWriteHex(U8 u8)
{
	UARTWriteHexNibble(u8>>4);
	UARTWriteHexNibble(u8);
}

static char UARTRead()
{
	while(!PIR1bits.RCIF)
	{
		NOP();
	}
	return RCREG;
}

static BOOL FlashIsBlank(void)
{
	// Check if Flash is all 0xFF
	BOOL bBlank=TRUE;
	
	U16 u16;
	
	for (u16=0;u16<NVRAM_SIZE && bBlank;u16++)
	{
		U8 u8;
		
		ReadFlash(u16+FLASH_OFFSET,1,&u8);
		if (u8!=0xFF)
		{
			bBlank=FALSE;
		}	
	}
	return bBlank;
}

static void FlashErase(void)
{
	UARTWriteText("\r\nErasing flash...\r\n");
	EraseFlash(FLASH_OFFSET,FLASH_OFFSET+NVRAM_SIZE-1);
}

static U16 FlashRandom(void)
{
	// Fill Flash with random data, returns checksum
	U16 u16;
	U8 au8[FLASH_WRITE_BLOCK];
	U16 u16Sum=0;
	
	FlashErase();
	
	for (u16=0;u16<NVRAM_SIZE;u16+=FLASH_WRITE_BLOCK)
	{
		U8 u8Count;
		U8 *pu8=au8;
		
		for (u8Count=0;u8Count<FLASH_WRITE_BLOCK;u8Count++)
		{
			U8 u8=(U8)rand();
			
			*pu8++=u8;
			u16Sum+=u8;
		}
		WriteBlockFlash(u16+FLASH_OFFSET,1,au8);
	}
	return u16Sum;
}

static U16 FlashChecksum(void)
{
	// Return simple checksum on flash
	U16 u16;
	U16 u16Sum=0;
	
	for (u16=0;u16<NVRAM_SIZE;u16++)
	{
		U8 u8;
		
		ReadFlash(u16+FLASH_OFFSET,1,&u8);
		u16Sum+=u8;
	}	
	return u16Sum;
}	

static void NVRAMInit(void)
{
	// Set up address, data and control lines for the NVRAM
	TRISD=0b11111111; // DQ lines, all inputs
	LAT_nCE=1;
	TRIS_nCE=0;
	LAT_nOE=1;
	TRIS_nOE=0;
	LAT_nWE=1;
	TRIS_nWE=0;
	TRISB &= 0b11000000; // RB0:RB5 -> A0:A5
	TRISA &= 0b10000000; // RA0:RA6 -> A6:A12
}

static void NVRAMSetAddr(U16 u16Addr)
{
	// Do appropriate bit shifting to get the 13 bit address out.
	LATB=(LATB & 0b11000000) | (u16Addr & 0b00111111);
	LATA=(LATA & 0b10000000) | ((u16Addr>>6) & 0b01111111);
}	

static U8 NVRAMRead(U16 u16Addr)
{
	// Read one byte of NVRAM
	U8 u8;
	
	LAT_nWE=1;
	LAT_nOE=1;
	LAT_nCE=1;
	TRISD=0b11111111;
	NVRAMSetAddr(u16Addr);
	LAT_nCE=0;
	LAT_nOE=0;
	u8=PORTD;
	LAT_nOE=1;
	LAT_nCE=1;
	
	return u8;
}	

static void NVRAMWrite(U16 u16Addr,U8 u8)
{
	// Write one byte of NVRAM
	LAT_nWE=1;
	LAT_nOE=1;
	LAT_nCE=1;
	LATD=u8;
	TRISD=0b00000000;
	NVRAMSetAddr(u16Addr);
	LAT_nCE=0;
	LAT_nWE=0;
	LAT_nWE=1;
	LAT_nCE=1;
	TRISD=0b11111111;
}

static void NVRAMErase(void)
{
	U16 u16;
	
	for (u16=0;u16<NVRAM_SIZE;u16++)
	{
		NVRAMWrite(u16,0xFF);
	}	
}	

static void NVRAMToFlash(void)
{
	U16 u16;
	U8 au8[FLASH_WRITE_BLOCK];
	
	FlashErase();
	
	for (u16=0;u16<NVRAM_SIZE;u16+=FLASH_WRITE_BLOCK)
	{
		U8 u8Count;
		U8 *pu8=au8;
		
		if ((u16 & 0x0FF)==0)
		{
			UARTWriteText("\rReading ");
			UARTWriteHex((U8)(u16>>8));
			UARTWriteHex((U8)u16);
		}
		for (u8Count=0;u8Count<FLASH_WRITE_BLOCK;u8Count++)
		{
			*pu8++=NVRAMRead(u16+u8Count);
		}
		WriteBlockFlash(u16+FLASH_OFFSET,1,au8);
	}
	UARTWriteText("\r\nRead completed.\r\n");
}	

static U16 NVRAMRandom(void)
{
	// Fill NVRAM with random data, returns checksum
	U16 u16;
	U16 u16Sum=0;
	
	for (u16=0;u16<NVRAM_SIZE;u16++)
	{
		U8 u8=(U8)rand();
			
		NVRAMWrite(u16,u8);
		u16Sum+=u8;
	}
	return u16Sum;
}


static U16 NVRAMChecksum(void)
{
	// Return simple checksum of NVRAM contents
	U16 u16;
	U16 u16Sum=0;
	
	for (u16=0;u16<NVRAM_SIZE;u16++)
	{
		U8 u8;
		
		u8=NVRAMRead(u16);
		u16Sum+=u8;
	}	
	return u16Sum;
}	

static BOOL Verify(void)
{
	// Compare flash and NVRAM contents
	U16 u16;
	BOOL bSuccess=TRUE;
	int nCount=0;
	
	for (u16=0;u16<NVRAM_SIZE && nCount<MAX_VERIFY_ERRORS;u16++)
	{
		U8 u8a;
		U8 u8b;
		
		ReadFlash(u16+FLASH_OFFSET,1,&u8a);

		if ((u16 & 0x0FF)==0)
		{
			UARTWriteText("\rVerifying ");
			UARTWriteHex((U8)(u16>>8));
			UARTWriteHex((U8)u16);
		}

		u8b=NVRAMRead(u16);

		if (u8a!=u8b)
		{
			UARTWriteText("\rVerify error: ");
			UARTWriteHex((U8)(u16>>8));
			UARTWriteHex((U8)u16);
			UARTWrite(' ');
			UARTWriteHex(u8a);
			UARTWrite(' ');
			UARTWriteHex(u8b);
			UARTWriteText("\r\n");
			bSuccess=FALSE;
			nCount++;
		}	
	}	
	UARTWriteText("\r\nVerify completed: ");
	if (bSuccess)
	{
		UARTWriteText("success.\r\n");
	}
	else
	{
		UARTWriteText("*** failed, maximum 10 errors shown. ***\r\n");
	}	
	return bSuccess;
}

static void FlashToNVRAM(void)
{
	// Copy PIC's flash to NVRAM
	U16 u16;
	
	for (u16=0;u16<NVRAM_SIZE;u16++)
	{
		U8 u8;
		
		ReadFlash(u16+FLASH_OFFSET,1,&u8);

		if ((u16 & 0x0FF)==0)
		{
			UARTWriteText("\rWriting ");
			UARTWriteHex((U8)(u16>>8));
			UARTWriteHex((U8)u16);
		}
		NVRAMWrite(u16,u8);
	}	
	UARTWriteText("\r\nWrite completed.\r\n");
}	

int main(void)
{
	U8 u8RCON=RCON; // Check what's reset us
	
	RCON=0b00011111;
	
	OSCCONbits.IRCF=0b111;
	
	NVRAMInit();
	UARTInit(115200L);
	
	UARTWriteText("\r\n");
	UARTWriteHex(u8RCON);
	UARTWriteText(": reset status\r\n");
	
	while (1)
	{
		char c;
		
		UARTWriteText("\r\n--- DS1225Y NVRAM copy programmer ---\r\n");
		UARTWriteText
		(
			"Select:\r\n"
			"  [R]ead NVRAM into PIC flash\r\n"
			"  [V]erify NVRAM with PIC flash\r\n"
			"  [W]rite PIC flash to NVRAM\r\n"
			"  erase PIC [F]lash\r\n"
			"  erase [N]VRAM\r\n"
			"  [B]lank check PIC flash\r\n"
			"  fill NVRAM with ran[D]om data\r\n"
			"  fill PIC flash with rand[O]m data\r\n"
			"  [P]rint NVRAM contents\r\n"
			"  print PIC f[L]ash contents\r\n"
			"  get [C]hecksums\r\n> "
		);
		c=toupper(UARTRead());
		if (c>=0x20 && c<=0x7E) // Avoid echoing control or non-printable characters (XON/XOFF for example!!!)
		{
			UARTWrite(c);
		}	
		UARTWriteText("\r\n");
		
		switch(c)
		{
			case 'B':
				// [B]lank check PIC flash
				if (FlashIsBlank())
				{
					UARTWriteText("Flash is blank.\r\n");
				}
				else
				{
					UARTWriteText("Flash is not blank.\r\n");
				}
				break;
			case 'C':
				// get [C]hecksums
				{
					U16 u16;
					
					u16=NVRAMChecksum();
					UARTWriteText("NVRAM checksum=0x");
					UARTWriteHex((U8)(u16>>8));
					UARTWriteHex((U8)u16);
					u16=FlashChecksum();
					UARTWriteText("\r\nFlash checksum=0x");
					UARTWriteHex((U8)(u16>>8));
					UARTWriteHex((U8)u16);
					UARTWriteText("\r\n");
				}
				break;							
			case 'D':
				// fill NVRAM with ran[D]om data
				{
					U16 u16=NVRAMRandom();

					UARTWriteText("NVRAM filled with random data.\r\n");
					UARTWriteText("Write checksum: 0x");
					UARTWriteHex((U8)(u16>>8));
					UARTWriteHex((U8)u16);
					u16=NVRAMChecksum();
					UARTWriteText("\r\nRead checksum:  0x");
					UARTWriteHex((U8)(u16>>8));
					UARTWriteHex((U8)u16);
					UARTWriteText("\r\n");
				}
				break;							
			case 'F':
				// erase PIC [F]lash
				FlashErase();
				UARTWriteText("Flash erased.\r\n");
				break;							
			case 'L':
				// print PIC f[L]ash contents
				{
					U16 u16;
					
					for (u16=0;u16<NVRAM_SIZE;u16++)
					{
						U8 u8;
						
						ReadFlash(u16+FLASH_OFFSET,1,&u8);
						
						if ((u16 & 0x0F)==0)
						{
							UARTWriteText("\r\n");
							UARTWriteHex((U8)(u16>>8));
							UARTWriteHex((U8)u16);
						}	
						UARTWrite(' ');
						UARTWriteHex(u8);
					}	
					UARTWriteText("\r\n");
				}
				break;							
			case 'N':
				// erase [N]VRAM
				NVRAMErase();
				UARTWriteText("NVRAM erased.\r\n");
				break;							
			case 'O':
				// fill PIC flash with rand[O]m data
				{
					U16 u16=FlashRandom();
					UARTWriteText("Flash filled with random data.\r\n");
					UARTWriteText("Write checksum: 0x");
					UARTWriteHex((U8)(u16>>8));
					UARTWriteHex((U8)u16);
					u16=FlashChecksum();
					UARTWriteText("\r\nRead checksum:  0x");
					UARTWriteHex((U8)(u16>>8));
					UARTWriteHex((U8)u16);
					UARTWriteText("\r\n");
				}
				break;							
			case 'P':
				// [P]rint NVRAM contents
				{
					U16 u16;
					
					for (u16=0;u16<NVRAM_SIZE;u16++)
					{
						U8 u8=NVRAMRead(u16);
						
						if ((u16 & 0x0F)==0)
						{
							UARTWriteText("\r\n");
							UARTWriteHex((U8)(u16>>8));
							UARTWriteHex((U8)u16);
						}	
						UARTWrite(' ');
						UARTWriteHex(u8);
					}	
					UARTWriteText("\r\n");
				}
				break;							
			case 'R':
				// [R]ead NVRAM into PIC flash
				NVRAMToFlash();
				Verify();
				break;
			case 'V':
				// [V]erify NVRAM with PIC flash
				if (FlashIsBlank())
				{
					UARTWriteText("*** Warning: Flash appears to be blank.***\r\n");
				}	
				Verify();
				break;
			case 'W':
				// [W]rite PIC flash to NVRAM
				if (FlashIsBlank())
				{
					UARTWriteText("*** Flash is blank, aborted. ***\r\n");
				}	
				else
				{
					FlashToNVRAM();
					Verify();
				}
				break;
			default:
				UARTWriteText("Unknown command\r\n");
				break;
		}	
	}	
	return 0;	
}

